/*
 *      vii - buffer and display output
 *      Copyright (C) 1991-1995, 1999, 2005, 2009, 2010 Peter Miller
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see
 *      <http://www.gnu.org/licenses/>.
 */

#include <lib/ac/string.h>
#include <lib/ac/stdio.h>
#include <lib/ac/stdlib.h>
#include <lib/ac/termios.h>
#include <libexplain/output.h>

#include <lib/option.h>
#include <lib/trace.h>


#define MIN_PAGE_WIDTH 40
/* #define MAX_PAGE_WIDTH in option.h */
#define DEFAULT_PAGE_WIDTH 80
#define MIN_PAGE_LENGTH 10
#define MAX_PAGE_LENGTH 30000
#define DEFAULT_PAGE_LENGTH 24

#define DEFAULT_TAB_WIDTH 8

#define LEVEL_TERSE 1
#define LEVEL_UNFORMATTED 2
#define LEVEL_UNSET 3
#define LEVEL_VERBOSE 4

static int      page_length;
static int      verbose_flag = LEVEL_UNSET;
static int      page_width;
static int      tab_width;


static void
too_many(void)
{
    explain_output_error_and_die
    (
        "only one of -TERse and -UNFormatted and -Verbose may be specified"
    );
}


void
option_verbose_set(void)
{
    trace(("option_set_verbose()\n{\n"));
    if (verbose_flag == LEVEL_VERBOSE)
        explain_output_error_and_die("duplicate -Verbose option");
    if (verbose_flag != LEVEL_UNSET)
        too_many();
    verbose_flag = LEVEL_VERBOSE;
    trace(("}\n"));
}


int
option_verbose_get()
{
    return (verbose_flag == LEVEL_VERBOSE);
}


void
option_terse_set(void)
{
    trace(("option_set_terse()\n{\n"));
    if (verbose_flag == LEVEL_TERSE)
        explain_output_error_and_die("duplicate -TERse option");
    if (verbose_flag != LEVEL_UNSET)
        too_many();
    verbose_flag = LEVEL_TERSE;
    trace(("}\n"));
}


int
option_terse_get(void)
{
    return (verbose_flag == LEVEL_TERSE);
}


void
option_unformatted_set(void)
{
    trace(("option_set_unformatted()\n{\n"));
    if (verbose_flag == LEVEL_UNFORMATTED)
        explain_output_error_and_die("duplicate -UNFormatted option");
    if (verbose_flag != LEVEL_UNSET)
        too_many();
    verbose_flag = LEVEL_UNFORMATTED;
    trace(("}\n"));
}


int
option_unformatted_get(void)
{
    return (verbose_flag == LEVEL_TERSE || verbose_flag == LEVEL_UNFORMATTED);
}


static void
default_page_sizes(void)
{
    if (!page_width)
    {
        const char      *cp;

        cp = getenv("COLS");
        if (cp)
        {
            int             n;

            n = atoi(cp);
            if (n > 0)
            {
                if (n < MIN_PAGE_WIDTH)
                    n = MIN_PAGE_WIDTH;
                if (n > MAX_PAGE_WIDTH)
                    n = MAX_PAGE_WIDTH;
                page_width = n;
            }
        }
    }
    if (!page_length)
    {
        const char      *cp;

        cp = getenv("LINES");
        if (cp)
        {
            int             n;

            n = atoi(cp);
            if (n > 0)
            {
                if (n < MIN_PAGE_LENGTH)
                    n = MIN_PAGE_LENGTH;
                if (n > MAX_PAGE_LENGTH)
                    n = MAX_PAGE_LENGTH;
                page_length = n;
            }
        }
    }

#ifdef TIOCGWINSZ
    if (!page_width || !page_length)
    {
        struct winsize ws;

        if (ioctl(0, TIOCGWINSZ, &ws) == 0)
        {
            if (!page_width && ws.ws_col > 0)
            {
                page_width = ws.ws_col;
                if (page_width < MIN_PAGE_WIDTH)
                    page_width = MIN_PAGE_WIDTH;
                if (page_width > MAX_PAGE_WIDTH)
                    page_width = MAX_PAGE_WIDTH;
            }
            if (!page_length && ws.ws_row > 0)
            {
                page_length = ws.ws_row;
                if (page_length < MIN_PAGE_LENGTH)
                    page_length = MIN_PAGE_LENGTH;
                if (page_length > MAX_PAGE_LENGTH)
                    page_length = MAX_PAGE_LENGTH;
            }
        }
    }
#endif

    if (!page_width)
        page_width = DEFAULT_PAGE_WIDTH;
    if (!page_length)
        page_length = DEFAULT_PAGE_LENGTH;
}


void
option_page_width_set(int n)
{
    if (page_width)
        explain_output_error_and_die("duplicate -Page_Width option");
    if (n < MIN_PAGE_WIDTH || n > MAX_PAGE_WIDTH)
        explain_output_error_and_die("page width %d out of range", n);
    page_width = n;
}


int
option_page_width_get(void)
{
    /*
     * must not generate a fatal error in this function,
     * as it is used by 'error.c' when reporting fatal errors.
     *
     * must not put tracing in this function,
     * because 'trace.c' uses it to determine the width.
     */
    if (!page_width)
        default_page_sizes();
    return page_width;
}


void
option_page_length_set(int n)
{
    if (page_length)
        explain_output_error_and_die("duplicate -Page_Length option");
    if (n < MIN_PAGE_LENGTH || n > MAX_PAGE_LENGTH)
        explain_output_error_and_die("page length %d out of range", n);
    page_length = n;
}


int
option_page_length_get(void)
{
    if (!page_length)
        default_page_sizes();
    return page_length;
}


void
option_tab_width_set(int n)
{
    if (n < 0 || n > 40)
        explain_output_error_and_die("tab width of %d is out of range", n);
    if (tab_width)
        explain_output_error_and_die("duplicate -Tab_Width option");
    if (n < 2)
        tab_width = -1;
    else
        tab_width = n;
}


int
option_tab_width_get(void)
{
    if (!tab_width)
        tab_width = DEFAULT_TAB_WIDTH;
    return (tab_width < 2 ? 0 : tab_width);
}
