/*
 *      vii - buffer and display output
 *      Copyright (C) 1991, 1995, 1999, 2005, 2009, 2010 Peter Miller
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see
 *      <http://www.gnu.org/licenses/>.
 */

#include <lib/ac/assert.h>
#include <lib/ac/stdlib.h>
#include <lib/ac/string.h>
#include <libexplain/malloc.h>
#include <libexplain/realloc.h>
#include <libexplain/strdup.h>

#include <vii/stash.h>
#include <vii/term.h>

/*
 * amount to increase line length by
 */
#define CHUNK   80


stash *
stash_open(void)
{
    char            *p;

    p = explain_malloc_or_die(sizeof(stash));
    memset(p, 0, sizeof(stash));
    return (stash *) p;
}


void
stash_free(stash *sp)
{
    int             j;

    for (j = 0; j < sp->nlines; ++j)
        free(sp->line[j]);
    free(sp);
}


void
stash_putc(stash *sp, int c)
{
    size_t          llen;

    assert(sp->x >= 0);
    assert(sp->y >= 0);

    /*
     * do control character processing
     */
    switch (c)
    {
    case '\r':
        sp->x = 0;
        return;

    case '\n':
        sp->y++;
        sp->x = 0;
        return;

    case '\t':
        sp->x = (sp->x + 8) & -8;
        return;

    case ' ':
        sp->x++;
        return;

    case '\b':
        if (sp->x > 0)
            sp->x--;
        return;
    }

    /*
     * cope with unprintable characters
     */
    if (c < ' ' || c > '~')
    {
        stash_putc(sp, '\\');
        stash_putc(sp, '0' + ((c >> 6) & 3));
        stash_putc(sp, '0' + ((c >> 3) & 7));
        c = '0' + (c & 7);
    }

    /*
     * don't stash if it gets too big
     */
    if (sp->x > 500 || sp->y > 5000)
        return;

    /*
     * get to the line
     */
    while (sp->y >= sp->nlines)
    {
        size_t nbytes;

        nbytes = (sp->nlines + 1) * sizeof(char *);
        sp->line = explain_realloc_or_die(sp->line, nbytes);
        sp->line[sp->nlines++] = explain_strdup_or_die("");
    }

    /*
     * see if the line is long enough
     */
    llen = strlen(sp->line[sp->y]);
    if (sp->x >= llen)
    {
        size_t          nbytes;

        nbytes = llen + CHUNK + 1;
        sp->line[sp->y] = explain_realloc_or_die(sp->line[sp->y], nbytes);
        memset(sp->line[sp->y] + llen, ' ', CHUNK);
        sp->line[sp->y][llen + CHUNK] = 0;
    }

    /*
     * place the character
     */
    sp->line[sp->y][sp->x++] = c;
}


void
stash_draw(stash *sp, int x, int y, int dx, int dy)
{
    int             col;
    int             line;

    assert(x >= 0);
    assert(y >= 0);
    for (line = y; line < sp->nlines && line < y + dy; ++line)
    {
        size_t          llen;

        llen = strlen(sp->line[line]);
        for (col = x; col < llen && col < x + dx; ++col)
        {
            term_putc(col - x, line - y, sp->line[line][col]);
        }
    }
}


int
stash_length(stash *sp)
{
    return sp->nlines;
}
