/*
 *  vii - buffer and display outout
 *  Copyright (C) 1991, 1992, 1995, 1999, 2003, 2005, 2009, 2010 Peter Miller
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <lib/ac/errno.h>
#include <lib/ac/signal.h>
#include <lib/ac/stdlib.h>
#include <lib/ac/unistd.h>
#include <lib/ac/poll.h>
#include <lib/ac/string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <curses.h>
#include <libexplain/output.h>
#include <libexplain/program_name.h>
#include <libexplain/read.h>

#include <lib/arglex.h>
#include <lib/option.h>

#include <vii/term.h>


static RETSIGTYPE
rubout(int n)
{
    const char      *progname;

    term_close();
    if (n && n != SIGINT)
    {
        progname = explain_program_name_get();
        psignal(n, progname);
#ifdef DEBUG
        abort();
#endif
    }
    exit(1);
}


static RETSIGTYPE
nop(int n)
{
    signal(n, nop);
}


#ifndef HAVE_ISENDWIN

static int
isendwin(void)
{
#ifndef HAVE__ENDWIN
    return 0;
#else
    return _endwin;
#endif
}

#endif


#ifdef  SIGSTOP

/*
 * What to do for stop interrupts.
 */

static RETSIGTYPE
stop(int n)
{
#if defined(HAVE_SIGPROCMASK)
    sigset_t        set;
#endif

    if (curscr && !isendwin())
    {
#ifndef HAVE_ISENDWIN
        /* clear the screen, if possible */
        fputs(CL, stdout);
#endif
        endwin();
    }
    fflush(stdout);

    signal(n, SIG_DFL);
#if defined(HAVE_SIGPROCMASK)
    sigemptyset(&set);
    sigaddset(&set, n);
    sigprocmask(SIG_BLOCK, &set, 0);
#elif defined(HAVE_SIGSETMASK)
    sigsetmask(sigblock(0) & ~(1 << (n - 1)));
#endif
    kill(0, n); /* will stop here until continued */
#if defined(HAVE_SIGPROCMASK)
    sigprocmask(SIG_UNBLOCK, &set, 0);
#elif defined(HAVE_SIGSETMASK)
    sigblock(1 << (n - 1));
#endif
    signal(n, stop);

    /* restore our terminal state */
    /* restore screen */
    if (curscr)
        wrefresh(curscr);
}

#endif


void
term_open(const char *ttype)
{
    int             j;

    if (!isatty(1) || !isatty(2))
        explain_output_error_and_die("not interactive");
    if (ttype)
    {
#ifndef HAVE_ISENDWIN
        Def_term = ttype;
        My_term = TRUE;
#endif
    }

    for (j = 1; j <= NSIG; j++)
    {
        switch (j)
        {
        default:
            if (signal(j, SIG_IGN) != SIG_IGN)
                signal(j, rubout);
            break;

#ifdef  SIGCHLD
        case SIGCHLD:
#endif
#ifdef  SIGTINT
        case SIGTINT:
#endif
#ifdef DEBUG
        case SIGSEGV:
        case SIGBUS:
        case SIGABRT:
#endif
        case SIGKILL:
            break;

#ifdef  SIGSTOP
        case SIGSTOP:
        case SIGCONT:
            break;

        case SIGTSTP:
        case SIGTTIN:
        case SIGTTOU:
            signal(j, stop);
            break;
#endif

        case SIGALRM:
            signal(j, nop);
            break;
        }
    }
    initscr();
    cbreak();
    noecho();
    leaveok(stdscr, FALSE);
}


void
term_close(void)
{
    if (stdscr && !isendwin())
    {
        move(LINES - 1, 0);
        clrtoeol();
        refresh();
        endwin();
    }
}


void
term_refresh(void)
{
    refresh();
}


void
term_clear(void)
{
    erase();
}


void
term_puts(int x, int y, const char *s)
{
    while (*s)
        term_putc(x++, y, *s++);
}


void
term_putc(int x, int y, int c)
{
    if (x < 0 || y < 0 || x >= COLS || y >= LINES)
        return;
    if (x == COLS - 1 && y == LINES - 1)
        return;
    if (c < ' ' || c > '~')
        c = '?';
    move(y, x);
    addch(c);
}


int
term_getc(int efd, void (*efunc)(int))
{
#ifdef HAVE_POLL
    for (;;)
    {
        struct pollfd   fds[2];
        int             nfds;

#ifndef POLLRDNORM
#define POLLRDNORM POLLIN
#endif
        fds[0].fd = 0;
        fds[0].events = POLLIN | POLLERR | POLLHUP | POLLRDNORM;
        fds[0].revents = 0;
        nfds = 1;

        if (efd >= 0 && efunc)
        {
            fds[1].fd = efd;
            fds[1].events = POLLIN | POLLERR | POLLHUP | POLLRDNORM;
            fds[1].revents = 0;
            nfds = 2;
        }

        if (poll(fds, nfds, -1) < 0)
        {
            if (errno == EINTR)
                return TERM_INTR;
            explain_output_error_and_die("poll: %s", strerror(errno));
        }

        if (fds[0].revents)
        {
            unsigned char   c;

            explain_read_or_die(0, &c, 1);
            return c;
        }
        if (nfds > 1 && fds[1].revents)
        {
            unsigned char   c;
            int             result;

            result = explain_read_or_die(efd, &c, 1);
            if (result == 0)
                return TERM_EOF;
            efunc(c);
        }
    }
#else
    for (;;)
    {
        fd_set          rfds;
        fd_set          wfds;
        fd_set          efds;
        int             result;

        FD_ZERO(&rfds);
        FD_ZERO(&wfds);
        FD_ZERO(&efds);
        FD_SET(0, &rfds);
        if (efd >= 0 && efunc)
            FD_SET(efd, &rfds);

        result = select(FD_SETSIZE, &rfds, &wfds, &efds, (struct timeval *)0);
        if (result < 0)
        {
            if (errno == EINTR)
                return TERM_INTR;
            nfatal("select");
        }
        if (FD_ISSET(0, &rfds))
        {
            unsigned char   c;

            if (read(0, &c, 1) != 1)
                nfatal("stdin");
            return c;
        }
        if (FD_ISSET(efd, &rfds))
        {
            unsigned char   c;

            result = read(efd, &c, 1);
            if (result < 0)
                nfatal("fd%d", efd);
            if (result == 0)
                return TERM_EOF;
            efunc(c);
        }
    }
#endif
}


void
term_last_line(void)
{
    move(LINES - 1, 0);
}


void
term_bell(void)
{
    putchar('\7');
    fflush(stdout);
}


void
term_redraw(void)
{
    wrefresh(curscr);
}


void
term_get_size(int *x, int *y)
{
    *x = COLS;
    *y = LINES;
}
