/*
 *      vii - buffer and display outout
 *      Copyright (C) 1991-1995, 1997, 1999, 2001, 2005, 2009, 2010 Peter Miller
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see
 *      <http://www.gnu.org/licenses/>.
 */

#include <lib/ac/stdio.h>
#include <lib/ac/string.h>
#include <lib/ac/time.h>
#include <lib/ac/unistd.h>
#include <libexplain/freopen.h>
#include <libexplain/pclose.h>
#include <libexplain/popen.h>

#include <lib/str.h>

#include <vii/veneer.h>
#include <vii/stash.h>
#include <vii/term.h>

#define CTRL(c) ((c) & 0x9F)

static int      quitting;
static int      pos_x;
static int      pos_y;
const char      *terminal;
int             sleep_time = 10;
static stash    *curpage;
static stash    *newpage;
static char     *show_time;
static int      show_position;


static void
update(void)
{
    int             x;
    int             y;
    int             show_row;

    if (!curpage)
        return;
    term_clear();
    term_get_size(&x, &y);
    stash_draw(curpage, pos_x, pos_y, x, y);

    show_row = 0;
    if (show_time)
    {
        char            buffer[80];
        time_t          now;
        struct tm       *tm;

        time(&now);
        tm = localtime(&now);
        buffer[0] = ' ';
        strftime(buffer + 1, sizeof(buffer) - 1, show_time, tm);
        term_puts(x - strlen(buffer), show_row, buffer);
        ++show_row;
    }

    if (show_position)
    {
        char            buffer[80];

        snprintf
        (
            buffer,
            sizeof(buffer),
            " Lines %d-%d of %d",
            pos_y + 1,
            pos_y + y,
            stash_length(curpage)
        );
        term_puts(x - strlen(buffer), show_row, buffer);
        ++show_row;
    }

    term_last_line();
    term_refresh();
}


static void
remember(int c)
{
    if (!newpage)
        return;
    if (c == '\f')
    {
        if (curpage)
            stash_free(curpage);
        curpage = newpage;
        newpage = stash_open();
        update();
    }
    else
        stash_putc(newpage, c);
}


static void
command(int c)
{
    int             x;
    int             y;
    char            *tmode;

    switch (c)
    {
    default:
        term_bell();
        break;

    case 'q':
    case 'Q':
        quitting = 1;
        break;

    case CTRL('L'):
    case CTRL('R'):
        term_redraw();
        break;

    case CTRL('E'):
    case 'j':
    case '+':
    case '\r':
    case '\n':
        pos_y++;
        update();
        break;

    case CTRL('Y'):
    case 'k':
    case '-':
        if (pos_y > 0)
            pos_y--;
        update();
        break;

    case 'J':
    case ' ':
    case CTRL('F'):
        term_get_size(&x, &y);
        pos_y += y - y / 4;
        update();
        break;

    case 'K':
    case CTRL('B'):
        term_get_size(&x, &y);
        pos_y -= y - y / 4;
        if (pos_y < 0)
            pos_y = 0;
        update();
        break;

    case 'C':
    case 'c':
    case CTRL('C'):
        tmode = "%c"; /* time and date */
        goto set_tmode;

    case 'D':
    case 'd':
    case CTRL('D'):
        tmode = "%x"; /* date w/o time */
        goto set_tmode;

    case 'T':
    case 't':
    case CTRL('T'):
        tmode = "%X"; /* time w/o date */
      set_tmode:
        if (show_time && !strcmp(show_time, tmode))
            show_time = 0;
        else
            show_time = tmode;
        update();
        break;

    case 'P':
    case 'p':
    case CTRL('P'):
        show_position = !show_position;
        update();
        break;
    }
}


static void
veneer_gather(int fd)
{
    int             c;

    newpage = stash_open();
    for (;;)
    {
        c = term_getc(fd, remember);
        if (c < 0)
        {
            if (curpage)
                stash_free(curpage);
            curpage = newpage;
            newpage = 0;
            update();
            break;
        }
        command(c);
        if (quitting)
            break;
    }
}


static void
veneer_cmdloop(void)
{
    int             c;

    for (;;)
    {
        c = term_getc(-1, (void (*)(int))0);
        if (c < 0)
            break;
        command(c);
        if (quitting)
            break;
    }
}


void
veneer(const char *cmd)
{
    if (cmd)
    {
        term_open(terminal);
        for (;;)
        {
            FILE            *fp;
            time_t          start;
            time_t          finish;
            long            elapsed;
            string_ty       *cmd2;

            cmd2 = str_format("( %s ) 2>&1 < /dev/null", cmd);

            time(&start);
            fp = explain_popen_or_die(cmd2->str_text, "r");
            str_free(cmd2);
            veneer_gather(fileno(fp));
            explain_pclose_or_die(fp);
            while (!quitting)
            {
                time(&finish);
                elapsed = finish - start;
                if (elapsed >= sleep_time)
                    break;
                alarm(sleep_time - elapsed);
                veneer_cmdloop();
                alarm(0);
            }
            if (quitting)
                break;
        }
        term_close();
    }
    else
    {
        int             fd;

        /*
         * make a copy of the stdin file descriptor,
         * and reopen stdin to the terminal.
         */
        fd = dup(0);
        explain_freopen_or_die("/dev/tty", "r", stdin);

        /*
         * consume all the input,
         * and then do commands from stdin
         */
        term_open(terminal);
        veneer_gather(fd);
        while (!quitting)
            veneer_cmdloop();
        term_close();
    }
}
